from __future__ import annotations

from dna_mutation_quantifier.core import *

# type hinting
import typing
from typing import List

# reading abi files
from Bio import SeqIO

# fasta parsing
from Bio import Seq
from Bio.SeqIO.FastaIO import FastaTwoLineParser

# general
import math
from collections import defaultdict

class Dna:
  @staticmethod
  def reverse_complement(sequence : str) -> str:
    return Seq.reverse_complement(sequence)

class SequencingLocus:
    def __init__(self, a, c, g, t):
        val_sum = a + c + g + t
        self.a = 0
        self.c = 0
        self.g = 0
        self.t = 0
        if val_sum == 0:
          return
        self.a = a / val_sum
        self.c = c / val_sum
        self.g = g / val_sum
        self.t = t / val_sum

    def __getitem__(self, key):
        if key == "A":
            return self.a
        elif key == "C":
            return self.c
        elif key == "G":
            return self.g
        elif key == "T":
            return self.t

    @staticmethod
    def get_complement(source_locus : SequencingLocus) -> SequencingLocus:
      return SequencingLocus(
          a = source_locus.t,
          c = source_locus.g,
          g = source_locus.c,
          t = source_locus.a
      )

class SangerTraceData:

  def __init__(self, a_data, c_data, g_data, t_data):
    self.data = {
        "A" : a_data,
        "C" : c_data,
        "G" : g_data,
        "T" : t_data,
    }
    self.a_data = a_data
    self.c_data = c_data
    self.g_data = g_data
    self.t_data = t_data

  def __getitem__(self, key):
    return self.data[key]


  def get_subset(self, start : int, end : int) -> SangerTraceData :
    s = slice(start, end-1)
    return SangerTraceData(self.data["A"][s], self.data["C"][s], self.data["G"][s], self.data["T"][s])

class NormalizedTrace:

    @staticmethod
    def from_abi_file(abi_file : str) -> NormalizedTrace:

        obj = NormalizedTrace()

        obj.is_reversed = False
        obj.file = abi_file

        record = SeqIO.read(abi_file, "abi")

        list(record.annotations.keys())
        list(record.annotations["abif_raw"].keys())

        order = record.annotations['abif_raw']['FWO_1']

        base_to_data = {
        order[0:1].decode(): "DATA9",
        order[1:2].decode(): "DATA10",
        order[2:3].decode(): "DATA11",
        order[3:4].decode(): "DATA12"
        }

        channels = ["DATA9", "DATA10", "DATA11", "DATA12"]
        trace = defaultdict(list)
        for c in channels:
            trace[c] = record.annotations["abif_raw"][c]

        indices = record.annotations["abif_raw"]["PLOC2"]
        primary_basecall = record.annotations["abif_raw"]["PBAS2"]
        quality = list(record.annotations["abif_raw"]["PCON2"])

        max_qual = max(quality)

        obj.quality_absolute_max = max_qual

        obj.peak_indices = indices
        obj.sequence = primary_basecall.decode()
        obj.quality = [q/max_qual for q in quality]

        seqlen = len(obj.sequence)

        obj.raw_data = SangerTraceData(
            list(trace[base_to_data["A"]]),
            list(trace[base_to_data["C"]]),
            list(trace[base_to_data["G"]]),
            list(trace[base_to_data["T"]]))

        complex_trace = []

        for index in range(seqlen):
            peak_index = indices[index]
            vA = obj.raw_data["A"][peak_index]
            vC = obj.raw_data["C"][peak_index]
            vG = obj.raw_data["G"][peak_index]
            vT = obj.raw_data["T"][peak_index]
            seqloc = SequencingLocus(vA,vC,vG,vT)
            complex_trace.append(seqloc)

        obj.complex_trace = complex_trace

        log(f"Imported trace with {len(complex_trace)} points",2)

        return obj

    def __getitem__(self, key):
        return self.complex_trace[key]

    def calc_match(self, search_seq, template_start_index):
        seqlen = len(search_seq)

        if template_start_index + len(search_seq) > len(self.sequence):
            return 0

        score=0
        for probe_index in range(seqlen):
            template_index = template_start_index + probe_index
            score+=self[template_index][search_seq[probe_index]]
        return score

    def calc_matches(self, search_seq):

        seqlen = len(self.complex_trace)
        scores=[self.calc_match(search_seq, index) for index in range(seqlen)]

        return scores

    def get_top_match(self, search_seq : str, min_score_pc : float = 0.5) -> int:

        matches = self.calc_matches(search_seq)

        top_score = max(matches)
        log(f"top_score is {top_score}",3)
        if top_score < min_score_pc:
          return -1
        else:
          return matches.index(top_score)

    def get_value(self, nucleotide : str, nt_index : int) -> int:
      data_index = self.peak_indices[nt_index]
      return self.raw_data[nucleotide][data_index]

    @staticmethod
    def reverse(source_nt : NormalizedTrace) -> NormalizedTrace:
      result_nt = NormalizedTrace()

      result_nt.is_reversed = True
      result_nt.file = source_nt.file
      result_nt.quality_absolute_max = source_nt.quality_absolute_max
      result_nt.quality = list(reversed(source_nt.quality))

      result_nt.sequence = Dna.reverse_complement(source_nt.sequence)

      min_index = 0
      max_index = len(source_nt.raw_data.a_data) - 1
      result_nt.peak_indices = list(reversed([min_index + (max_index-ind) for ind in source_nt.peak_indices]))

      result_nt.raw_data = SangerTraceData(
          a_data = list(reversed(source_nt.raw_data.t_data)),
          c_data = list(reversed(source_nt.raw_data.g_data)),
          g_data = list(reversed(source_nt.raw_data.c_data)),
          t_data = list(reversed(source_nt.raw_data.a_data)))

      result_nt.complex_trace = list(reversed([SequencingLocus.get_complement(locus) for locus in source_nt.complex_trace]))

      return result_nt

